{%- extends "basic/layout.html" %}
{%- block extrahead %}
  {{ super() }}
  <meta name="viewport" content="width=device-width, initial-scale=1">
{% endblock %}

{% block relbar1 %}{% endblock %}
{%- block rootrellink %}{% endblock %}
{% block relbar2 %}
  {%- if next %}
   <div class="related" role="navigation" aria-label="related navigation">
    Next: <a href="{{ next.link|e }}">{{ next.title|striptags|e }}</a>
   </div>
  {%- endif %}
{% endblock %}

{%- block footer %}
  <div class="footer" role="contentinfo">
    Updated on {{ last_updated }} with Sphinx {{ sphinx_version }}.
    {%- if show_source and has_source and sourcename %}
      | <a href="{{ pathto('_sources/' + sourcename, true)|e }}"
          rel="nofollow">{{ _('Page source') }}</a>
    {%- endif %}
    {%- if theme_github_blob %}
        | <a href="https://github.com/{{ theme_github_blob }}/{{pagename}}.rst"
          rel="nofollow">{{ 'Edit on GitHub' }}</a>
    {%- endif %}
  </div>
{%- endblock %}
